<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Command;

use Dedi\SyliusMigrationPlugin\Dto\MigrationOption;
use Dedi\SyliusMigrationPlugin\Service\MigrationExecutorInterface;
use Exception;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;

class MigrateCommand extends Command
{
    private MigrationExecutorInterface $migrationExecutor;

    public function __construct(
        string $name = null,
        MigrationExecutorInterface $migrator
    ) {
        parent::__construct($name);
        $this->migrationExecutor = $migrator;
    }

    protected function configure(): void
    {
        $this
            ->setDescription('Migrate database from a Legacy Database to Sylius')
            ->addArgument('tables', InputArgument::OPTIONAL, 'List of target tables to import, comma separated', '')
            ->addOption('dry-run', null, InputOption::VALUE_NONE, 'Only execute a few lines per table and prints statement instead of executing them')
            ->addOption('truncate-tables', 't', InputOption::VALUE_NONE, 'Truncate configured tables before migrate')
            ->addOption('date-from', null, InputOption::VALUE_REQUIRED, 'Select only data from a specific date')
            ->addOption('group', null, InputOption::VALUE_REQUIRED, 'A group of tables to migrate')
        ;
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        /** @var bool $dryRun */
        $dryRun = $input->getOption('dry-run');
        /** @var bool $truncate */
        $truncate = $input->getOption('truncate-tables');
        /** @var string $tables */
        $tables = $input->getArgument('tables');
        /** @var string|null $dateFrom */
        $dateFrom = $input->getOption('date-from');
        /** @var string|null $group */
        $group = $input->getOption('group');

        if (true === $dryRun) {
            $output->writeln('<info>Running in dry-mode. No statement will be executed on databases</info>');
        }
        if (true === $truncate) {
            $output->writeln('<info>Tables will be truncated</info>');
        }
        if (null !== $dateFrom) {
            $dateFrom = (new \DateTime($dateFrom))->setTime(0, 0);
            $output->writeln('<info>Import data from ' . $dateFrom->format('d/m/Y') . '</info>');
        }

        try {
            $executionStartTime = microtime(true);
            $migrationOption = new MigrationOption($dryRun, $truncate, '' === $tables ? null : explode(',', $tables), $dateFrom, $group);
            $this->migrationExecutor->execute($migrationOption);

            $executionEndTime = microtime(true);
            $seconds = round($executionEndTime - $executionStartTime);
            $hours = round($seconds / 3600);
            $mins = round($seconds / 60 % 60);
            $secs = round($seconds % 60);
            $output->writeln(sprintf('<info>Migration terminée en %02d:%02d:%02d</info>', $hours, $mins, $secs));

            return Command::SUCCESS;
        } catch (Exception $e) {
            $output->writeln(sprintf('<error>%s</error>', $e->getMessage()));
        }

        return Command::FAILURE;
    }
}
