<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Repository;

use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Doctrine\DBAL\Connection;
use Doctrine\DBAL\Query\QueryBuilder;

class MappingRepository
{
    private MigrationConfig $config;

    private Connection $connection;

    public function __construct(
        Connection $connection,
        MigrationConfig $config
    ) {
        $this->connection = $connection;
        $this->config = $config;
    }

    public function save(TableMappingDefinition $tableMappingDefinition, array $data): void
    {
        foreach ($tableMappingDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
            $destColumnName = $identifierMappingDefinition->getDestColumnName();
            if ($destColumnName === 'id') {
                $destColumnName = 'dest_id';
            }
            $qb = $this->connection->createQueryBuilder();
            $qb->insert($this->config->getTempMappingTableName());

            $this->setSourceId($qb, $data[$identifierMappingDefinition->getSourceColumnName()]);
            $this->setDestId($qb, $data[$destColumnName]);

            $this->setSourceTableName($qb, $identifierMappingDefinition->getSourceTableName());
            $this->setSourceColumnName($qb, $identifierMappingDefinition->getSourceColumnName());

            $this->setDestTableName($qb, $identifierMappingDefinition->getDestTableName());
            $this->setDestColumnName($qb, $identifierMappingDefinition->getDestColumnName());

            $qb->execute();
        }
    }

    public function getAll(
        string $sourceTableName,
        string $sourceColumnName,
        ?string $destTableName = null,
        ?string $destColumnName = null,
        $sourceId = null,
        $destId = null
    ) {
        $qb = $this->connection->createQueryBuilder();
        $qb
            ->select('*')
            ->from($this->config->getTempMappingTableName())
        ;

        $this->whereSourceTableName($qb, $sourceTableName);
        $this->whereSourceColumnName($qb, $sourceColumnName);

        if (null !== $destTableName) {
            $this->whereDestTableName($qb, $destTableName);
        }

        if (null !== $destColumnName) {
            $this->whereDestColumnName($qb, $destColumnName);
        }

        if (null !== $sourceId) {
            $this->whereSourceId($qb, $sourceId);
        }

        if (null !== $destId) {
            $this->whereDestId($qb, $destId);
        }

        return $qb->execute()->fetchAllAssociative();
    }

    public function getOne(string $sourceTableName, string $sourceColumnName, string $sourceId, ?string $destTableName = null)
    {
        $qb = $this->connection->createQueryBuilder();
        $qb
            ->select('*')
            ->from($this->config->getTempMappingTableName())
            ->setMaxResults(1)
        ;

        $this->whereSourceTableName($qb, $sourceTableName);
        $this->whereSourceColumnName($qb, $sourceColumnName);
        $this->whereSourceId($qb, $sourceId);

        if (null !== $destTableName) {
            $this->whereDestTableName($qb, $destTableName);
        }

        return $qb->execute()->fetchAssociative();
    }

    public function getMappingIdentifiers()
    {
        $qb = $this->connection->createQueryBuilder();
        $qb
            ->select('source_table_name, source_column_name, dest_table_name, dest_column_name')
            ->from($this->config->getTempMappingTableName())
            ->distinct()
        ;

        return $qb->execute()->fetchAll();
    }

    public function countIdentifiers(string $sourceTableName, string $sourceColumnName, ?string $destTableName = null): int
    {
        $qb = $this->connection->createQueryBuilder();
        $qb
            ->select('count(id)')
            ->from($this->config->getTempMappingTableName())
        ;

        $this->whereSourceTableName($qb, $sourceTableName);
        $this->whereSourceColumnName($qb, $sourceColumnName);

        if (null !== $destTableName) {
            $this->whereDestTableName($qb, $destTableName);
        }

        return $qb->execute()->fetchOne();
    }

    public function setSourceTableName(QueryBuilder $qb, string $sourceTableName): void
    {
        $qb
            ->setValue('source_table_name', ':source_table_name')
            ->setParameter('source_table_name', $sourceTableName)
        ;
    }

    private function setDestId(QueryBuilder $qb, $destId): void
    {
        $qb
            ->setValue('dest_id', ':dest_id')
            ->setParameter('dest_id', $destId)
        ;
    }

    private function setSourceId(QueryBuilder $qb, $sourceId): void
    {
        $qb
            ->setValue('source_id', ':source_id')
            ->setParameter('source_id', $sourceId)
        ;
    }

    private function setDestColumnName(QueryBuilder $qb, string $destColumnName): void
    {
        $qb
            ->setValue('dest_column_name', ':dest_column_name')
            ->setParameter('dest_column_name', $destColumnName)
        ;
    }

    private function setSourceColumnName(QueryBuilder $qb, string $sourceColumnName): void
    {
        $qb
            ->setValue('source_column_name', ':source_column_name')
            ->setParameter('source_column_name', $sourceColumnName)
        ;
    }

    public function setDestTableName(QueryBuilder $qb, string $destTableName): void
    {
        $qb
            ->setValue('dest_table_name', ':destTable')
            ->setParameter('destTable', $destTableName)
        ;
    }

    public function whereSourceTableName(QueryBuilder $qb, string $sourceTableName): void
    {
        $qb
            ->andWhere('source_table_name = :source_table_name')
            ->setParameter('source_table_name', $sourceTableName)
        ;
    }

    public function whereDestTableName(QueryBuilder $qb, string $destTableName): void
    {
        $qb
            ->andWhere('dest_table_name = :destTable')
            ->setParameter('destTable', $destTableName)
        ;
    }

    private function whereDestId(QueryBuilder $qb, $destId): void
    {
        $qb
            ->andWhere('dest_id = :dest_id')
            ->setParameter('dest_id', $destId)
        ;
    }

    private function whereSourceId(QueryBuilder $qb, $sourceId): void
    {
        $qb
            ->andWhere('source_id = :source_id')
            ->setParameter('source_id', $sourceId)
        ;
    }

    private function whereDestColumnName(QueryBuilder $qb, string $destColumnName): void
    {
        $qb
            ->andWhere('dest_column_name = :dest_column_name')
            ->setParameter('dest_column_name', $destColumnName)
        ;
    }

    private function whereSourceColumnName(QueryBuilder $qb, string $sourceColumnName): void
    {
        $qb
            ->andWhere('source_column_name = :source_column_name')
            ->setParameter('source_column_name', $sourceColumnName)
        ;
    }

    public function delete(
        string $sourceTableName,
        string $sourceColumnName,
        string $destTableName,
        string $destColumName,
        $sourceId,
        $destId
    ) {
        $qb = $this->connection->createQueryBuilder();
        $qb->delete($this->config->getTempMappingTableName());

        $this->whereSourceId($qb, $sourceId);
        $this->whereDestId($qb, $destId);

        $this->whereSourceTableName($qb, $sourceTableName);
        $this->whereSourceColumnName($qb, $sourceColumnName);

        $this->whereDestTableName($qb, $destTableName);
        $this->whereDestColumnName($qb, $destColumName);

        $qb->execute();
    }
}
